using Test
using StaticArrays
using Proj
import PROJ_jll
using NetworkOptions: ca_roots

function read_cmd(cmd)
    bytes = read(cmd)
    str = String(bytes)
    # on Windows there are CFLR line endings
    lf_only = replace(str, "\r" => "")
    no_tabs = replace(lf_only, "\t" => " ")
    return String(strip(no_tabs))
end

# https://proj.org/apps/cs2cs.html#using-epsg-crs-codes
@test PROJ_jll.cs2cs() do cs2cs
    read_cmd(pipeline(IOBuffer("45N 2E"), `$cs2cs EPSG:4326 EPSG:32631`))
end == "421184.70 4983436.77 0.00"

# https://proj.org/usage/transformation.html?#grid-based-datum-adjustments
@test PROJ_jll.cs2cs() do cs2cs
    withenv("PROJ_NETWORK" => "ON", "PROJ_CURL_CA_BUNDLE" => ca_roots()) do
        read_cmd(
            pipeline(
                IOBuffer("-111 50"),
                `$cs2cs +proj=latlong +ellps=clrk66 +nadgrids=ca_nrc_ntv1_can.tif
                    +to +proj=latlong +ellps=GRS80 +datum=NAD83`,
            ),
        )
    end
end == "111d0'3.006\"W 50d0'0.125\"N 0.000"

function xyzt_transform_cli(point::AbstractVector; network::Bool = false)
    # convert the vector to a string like "-33 151 5 2020"
    input = String(strip(string(point'), ('[', ']')))
    xyzt_transform_cli(input; network)
end

function xyzt_transform_cli(point::String; network::Bool = false)
    proj_network = network ? "ON" : "OFF"
    PROJ_jll.cs2cs() do cs2cs
        withenv("PROJ_NETWORK" => proj_network) do
            read_cmd(pipeline(IOBuffer(point), `$cs2cs -d 6 EPSG:4326+5773 EPSG:7856+5711`))
        end
    end
end

function xyzt_transform(point::AbstractVector; network::Bool = false)
    proj_network = network ? "ON" : "OFF"
    withenv("PROJ_NETWORK" => proj_network) do
        trans = Proj.Transformation("EPSG:4326+5773", "EPSG:7856+5711", always_xy = true)
        trans(point)
    end
end

# http://osgeo-org.1560.x6.nabble.com/PROJ-Different-results-from-cs2cs-vs-C-API-code-td5446396.html

# When you use cs2cs and do not specify it (as the 4th value in the coordinate),
# the time is sent to HUGE_VAL, which makes it not being used at all
# (so here as the reference epoch for the time-dependent Helmert transformation is 2020.0,
# it will be as if you specified 2020.0).

xy1 = "313153.228163 6346937.876336"
xy2 = "313153.271257 6346937.914865"
@test xyzt_transform_cli(SA[-33, 151, 5, 2020]; network = true) == "$xy1 5.280603 2020"
@test xyzt_transform_cli("-33 151 5 2020"; network = true) == "$xy1 5.280603 2020"
# no z correction with no network, as expected
@test xyzt_transform_cli(SA[-33, 151, 5, 2020]; network = false) == "$xy2 5.000000 2020"
# year 0, different xyz, as expected
@test xyzt_transform_cli(SA[-33, 151, 5, 0]; network = true) ==
      "313153.228163 6346937.876336 5.280603 0"
# no time input is like passing 2020, as expected
@test xyzt_transform_cli(SA[-33, 151, 5]; network = true) == "$xy1 5.280603"

# need to switch axis order here
# interestingly this does not respond to the PROJ_NETWORK environment variable,
# but does respond to enabling it in the context
Proj.proj_context_is_network_enabled()
@test Proj.proj_context_set_enable_network(1) == 1
# this is like passing floatmax() to the cli, but if we do it correctly we expect
# it to be like the version that passes only xyz to the cli
p = xyzt_transform(SA_F64[151, -33, 5])
@test is_approx(p, (313153.2281628684, 6.346937876336246e6, 5.280603319143665))
# this is expected, like the above
p = xyzt_transform(SA_F64[151, -33, 5, 2020])
@test is_approx(p, (313153.2281628684, 6.346937876336246e6, 5.280603319143665, 2020.0))
p = xyzt_transform(SA_F64[151, -33, 5, floatmax()*2])
@test is_approx(p, (313153.2281628684, 6.346937876336246e6, 5.280603319143665, Inf))

PROJ_jll.cs2cs() do cs2cs
    withenv("PROJ_NETWORK" => "OFF") do
        @test read_cmd(
            pipeline(IOBuffer("-33 151 5 0"), `$cs2cs -d 8 EPSG:4326+5773 EPSG:7856+5711`),
        ) == "313153.27125690 6346937.91486492 5.00000000 0"
    end
    withenv("PROJ_NETWORK" => "ON") do
        @test read_cmd(
            pipeline(IOBuffer("-33 151 5 0"), `$cs2cs -d 8 EPSG:4326+5773 EPSG:7856+5711`),
        ) == "313153.22816287 6346937.87633625 5.28060332 0"
    end
end

@test PROJ_jll.projinfo() do projinfo
    read_cmd(`$projinfo  -o PROJ EPSG:25832`)
end == """PROJ.4 string:
+proj=utm +zone=32 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs +type=crs"""

PROJ_jll.projinfo() do projinfo
    n_candidate_ops = "Candidate operations found: 3"
    grid_not_found =
        "Grid us_nga_egm96_15.tif needed but not found on the system." *
        " Can be obtained at https://cdn.proj.org/us_nga_egm96_15.tif"
    info = read_cmd(`$projinfo -s EPSG:4326+5773 -t EPSG:7856+5711`)
    @test occursin(n_candidate_ops, info)
    @test occursin(grid_not_found, info)
end
